
def get_ip():
    import subprocess
    return subprocess.check_output('dig +short myip.opendns.com @resolver1.opendns.com'.split()).strip().decode('ascii')
    # import socket
    # sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    # sock.connect(('resolver1.opendns.com', 53))
    # sock.send(b'\0\0\1\0\0\1\0\0\0\0\0\0\4myip\7opendns\3com\0\0\1\0\1')
    # resp = sock.recv(1000)
    # return '.'.join(str(b) for b in resp[-4:])
    # import requests, re
    # data = requests.get('http://checkip.dyndns.com/').text
    # return re.compile(r'Address: (\d+\.\d+\.\d+\.\d+)').search(data).group(1)


def open_browser(url):
    import os
    import webbrowser
    if 'DISPLAY' not in os.environ:
        print('The DISPLAY variable is not set, so not attempting to open a web browser\n')
        return False
    for name in 'windows-default macosx chrome chromium mozilla firefox opera safari'.split():
        # Note: `macosx` fails on macOS 10.12.5 due to <http://bugs.python.org/issue30392>.
        try:
            b = webbrowser.get(name)
            if b.open(url):
                return True
        except Exception:
            pass
    return False


def run_gunicorn(app, host='0.0.0.0', port=5000, use_reloader=True, num_workers=4, accesslog='-'):
    '''takes a flask app or perhaps other kinds of wsgi apps'''
    # TODO: figure out how to suppress sigwinch
    import gunicorn.app.base
    class StandaloneGunicornApplication(gunicorn.app.base.BaseApplication):
        # from <http://docs.gunicorn.org/en/stable/custom.html>
        def __init__(self, app, opts=None):
            self.application = app
            self.options = opts or {}
            super().__init__()
        def load_config(self):
            for key, val in self.options.items():
                self.cfg.set(key, val)
        def load(self):
            return self.application
    options = {
        'bind': '{}:{}'.format(host, port),
        'reload': use_reloader,
        'workers': num_workers,
        'accesslog': accesslog,
        'access_log_format': '%(t)s | %(s)s | %(L)ss | %(m)s %(U)s %(q)s | resp_len:%(B)s | referrer:"%(f)s" | ip:%(h)s | agent:%(a)s',
        # docs @ <http://docs.gunicorn.org/en/stable/settings.html#access-log-format>
        'worker_class': 'gevent',
    }
    sga = StandaloneGunicornApplication(app, options)
    # # debugging:
    # for skey,sval in sorted(sga.cfg.settings.items()):
    #     cli_args = sval.cli and ' '.join(sval.cli) or ''
    #     val = str(sval.value)
    #     print(f'cfg.{skey:25} {cli_args:28} {val}')
    #     if sval.value != sval.default:
    #         print(f'             default: {str(sval.default)}')
    #         print(f'             short: {sval.short}')
    #         print(f'             desc: <<\n{sval.desc}\n>>')
    sga.run()


from .func_cache_utils import shelve_cache
@shelve_cache
def cached_get(url):
    import requests
    # Do we need to encode this response somehow?
    # Maybe as `(resp.status, resp.text)`?
    headers = {'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.13; rv:62.0) Gecko/20100101 Firefox/62.0',
               'Accept-Language':'en-US,en;q=0.5',
    }
    return requests.get(url, headers=headers)



import itertools, functools
from boltons import iterutils

def Fn(s):
    return eval('(lambda _: ({}))'.format(s.replace('`',"'''")))
def Fn_idx(s):
    return eval('(lambda _,idx: ({}))'.format(s.replace('`',"'''")))
def Fn2(s):
    return eval('(lambda _,__: ({}))'.format(s.replace('`',"'''")))

def try_except_pass(func, exceptions=Exception):
    if isinstance(func, str): func = Fn(func)
    try: return func()
    except exceptions: return None

class KpaIterable:
    def __init__(self, iterable): self._iterable = iterable
    def __iter__(self): return iter(self._iterable)
    def __repr__(self):
        lst = self.take(1001)
        if len(lst) <= 10: return '<KpaIterable len={} values={}>'.format(len(lst), lst)
        elif len(lst) <= 1000: return '<KpaIterable len={} values={}...>'.format(len(lst), lst[:10])
        else: return '<KpaIterable len={}+ values={}...>'.format(1000, lst[:10])

    # map-like
    def map(self, func):
        if isinstance(func, str): func = Fn_idx(func)
        two_arg = try_except_pass(lambda: func.__code__.co_argcount == 2, AttributeError)
        if two_arg: return KpaIterable(func(item,idx) for idx,item in enumerate(self))
        else: return KpaIterable(func(item) for item in self)
    def mmap(self, func):
        if isinstance(func, str): func = Fn_idx(func)
        two_arg = try_except_pass(lambda: func.__code__.co_argcount == 2, AttributeError)
        if two_arg: return KpaIterable(itertools.chain.from_iterable(func(item,idx) for idx,item in enumerate(self)))
        else: return KpaIterable(itertools.chain.from_iterable(func(item) for item in self))
    def filter(self, func):
        if isinstance(func, str): func = Fn_idx(func)
        two_arg = try_except_pass(lambda: func.__code__.co_argcount == 2, AttributeError)
        if two_arg: return KpaIterable(item for idx,item in enumerate(self) if func(item,idx))
        else: return KpaIterable(item for item in self if func(item))

    # to one item
    def reduce(self, func, *initial): func = Fn2(func) if isinstance(func,str) else func; return functools.reduce(func, self, *initial)
    def join(self, s=';'): return s.join(self.map(str))
    def len(self): return sum(1 for _ in self)
    def all(self): return all(self)
    def any(self): return any(self)
    def all_same(self): return iterutils.same(self)

    # to list
    def head(self, n=10): return KpaIterable(item for _,item in zip(range(n), self))
    def skip(self, n): return KpaIterable(item for idx,item in enumerate(self) if idx>=n)
    def tail(self, n=10):
        if n == 0: return KpaIterable([])
        ret = [None]*n
        for idx, item in enumerate(self): ret[idx%n] = item
        return KpaIterable(ret[:idx+1] if idx<n else ret[idx+1:] + ret[:idx+1])
    def list(self): return list(self._iterable) # trivial if isinstance(self._iterable, list)
    def take(self, n=5): return self.head(n).list()

    # fancy
    def uniq(self): return KpaIterable(iterutils.unique_iter(self)) # uses set(), so not quite general
    def sorted(self, key=None): return KpaIterable(sorted(self, key=Fn(key) if isinstance(key,str) else key))
    def chunks(self, size): return KpaIterable(iterutils.chunked_iter(self, size))
    def windows(self, size): return KpaIterable(iterutils.windowed_iter(self, size))
    def flatten(self): return KpaIterable(itertools.chain.from_iterable(KpaIterable(item).flatten() if isinstance(item, (list,tuple)) else [item] for item in self))
    def _accumulate(self):
        for idx,item in enumerate(self):
            if idx == 0: total = item
            else: total += item
            yield total
    def accumulate(self): return KpaIterable(self._accumulate())
    def diffs(self): return self.windows(2).map('_[1] - _[0]') # I wish this included the first item


if __name__ == '__main__':
    # TODO: make these run via pytest
    assert Fn('_ % 30')(10) == 10

    repr(KpaIterable(itertools.count()))

    # map-like
    KpaIterable(itertools.count()).map('_*2').take()
    KpaIterable(itertools.count()).mmap('[9-_]*_').take()
    KpaIterable(itertools.count()).filter('_*2 != 4').take()

    # to one item
    KpaIterable(range(100)).reduce('_ + __')
    KpaIterable(range(100)).reduce('_ + __', -1000)
    KpaIterable(range(100)).map('[str(_)]').reduce('_ + __')
    KpaIterable(range(100)).map('[str(_)]').reduce('_ + __', [7])
    KpaIterable(range(100)).join()
    KpaIterable(range(100)).join(' ')
    KpaIterable(range(100)).len()
    KpaIterable(range(100)).all()
    KpaIterable(itertools.count()).any()
    KpaIterable(range(100)).map('[_,str(_),{(_):{_}}]').all_same()

    # to list
    KpaIterable(itertools.count()).head(0)
    KpaIterable(itertools.count()).head(1)
    KpaIterable(range(100)).head(1000)
    KpaIterable(range(100)).tail(0)
    KpaIterable(range(100)).tail(1)
    KpaIterable(range(100)).tail(1000)
    KpaIterable(itertools.count()).skip(1000).take()
    assert KpaIterable(range(100)).skip(0).list() == KpaIterable(range(100)).list()
    assert KpaIterable(range(100)).skip(10).list() == KpaIterable(range(100)).list()[10:]
    assert KpaIterable(range(100)).skip(1000).list() == []

    # fancy
    KpaIterable(itertools.count()).uniq().take()
    KpaIterable(itertools.count()).head().sorted()
    KpaIterable(itertools.count()).chunks(2).take()
    KpaIterable(itertools.count()).windows(2).take()
    KpaIterable(itertools.count()).chunks(2).flatten().take()
    KpaIterable(itertools.count()).accumulate().take()
    KpaIterable(itertools.count()).diffs().take()

    # etc
    KpaIterable(itertools.count()).map('_*2').mmap('[str(9-_)]*_').windows(2).take()
    KpaIterable(itertools.count()).map('_*2').mmap('[str(9-_)]*_').windows(2).flatten().head().sorted().uniq().map(type).tail().join()
    KpaIterable(itertools.count()).map('_*2').mmap('[str(9-_)]*_').windows(2).flatten().head().sorted().map(int).reduce('_ + __')



#!/usr/bin/env python3

def main():
    import sys, pathlib
    from .func_utils import assign

    if sys.argv[1:] == ['status-code-server']:
        from .http_server import serve, status_code_server
        serve(status_code_server)

    elif sys.argv[1:] == ['dir-server']:
        from .http_server import serve, magic_dir_server
        serve(magic_dir_server)

    elif sys.argv[1:] and sys.argv[1] == 'redirect-server':
        from .http_server import serve, make_redirect_server
        port = int(sys.argv[2])
        target_base_url = sys.argv[3]
        serve(make_redirect_server(target_base_url), port=port)

    elif sys.argv[1:] == ['termcolor']:
        from .terminal_utils import termcolor
        def r(num): return '#' if num is None else str(num%10)
        print('    # ' + ' '.join('{bg:2}'.format(bg=bg) for bg in range(50)))
        for fg in [None]+list(range(0,25)):
            print('{fg if fg is not None else "#":<2} '.format(fg=fg) +
                  ' '.join(termcolor(r(fg)+r(bg), fg, bg) for bg in [None]+list(range(50))))

    elif sys.argv[1:] and sys.argv[1] == "pip-find-updates":
        from .pip_utils import check_file
        @assign
        def filepath():
            if sys.argv[2:]:
                p = pathlib.Path(sys.argv[2])
                return p.as_posix() if p.exists() else None
            p = pathlib.Path().absolute()
            for directory in [p] + list(p.parents):
                for filename in ['setup.py', 'requirements.txt']:
                    p = directory / filename
                    if p.exists(): return p.as_posix()
            return None
        if not filepath: print('argv[2] not found and failed to find setup.py or requirements.txt in parent dirs')
        else:
            print('Looking at {filepath}'.format(filepath=filepath))
            check_file(filepath)

    else:
        if sys.argv[1:]: print('unknown command:', sys.argv[1:])
        print('available commands:\n  kpa termcolor\n  kpa status-code-server\n  kpa redirect-server\n  kpa pip-find-updates')




def run(filepath:str = ''):
    run_flake8(filepath)
    run_mypy(filepath)

def run_flake8(filepath:str = ''):
    import subprocess, sys
    p = subprocess.run(['flake8', '--show-source', '--ignore=E501,E302,E251,E701,E226,E305,E225,E261,E231,E301,E306,E402,E704,E265,E201,E202,E303,E124,E241,E127,E266,E221,E126,E129,F811,E222,E401,E702,E203,E116,E228,W504,W293,B007,W391', filepath])
    if p.returncode != 0: sys.exit(1)

def run_mypy(filepath:str = ''):
    import subprocess, sys
    p = subprocess.run(['mypy', '--pretty', '--ignore-missing-imports'] + ([filepath] if filepath else []))
    if p.returncode != 0: sys.exit(1)




def _f():
    import sys
    if sys.version_info < (3, 4):
        print("Requires Python 3.4+")
        sys.exit(1)

    import math
    try: math.inf
    except AttributeError: math.inf = float('inf')
_f()
del _f



#!/usr/bin/env python3
# to install locally: `pip install -e .`
# to install latest from pypi: `pip3 install --upgrade --upgrade-strategy eager --no-cache-dir kpa`
# to publish: `./setup.py publish`
# to update deps: `kpa pip-find-updates`
# to test: `./setup.py test`

from setuptools import setup
import importlib
import sys
from pathlib import Path
from urllib.request import urlopen
import subprocess
import json

def load_module_by_path(module_name, filepath):
    module = importlib.util.module_from_spec(importlib.util.spec_from_file_location(module_name, filepath))
    module.__spec__.loader.exec_module(module)
    return module
version = load_module_by_path('kpa.version', 'kpa/version.py').version


if sys.argv[-1] in ['publish', 'pub']:
    # TODO: use `class UploadCommand(setuptools.Command)` from <https://github.com/kennethreitz/setup.py/blob/master/setup.py#L49>

    git_workdir_returncode = subprocess.run('git diff-files --quiet'.split()).returncode
    assert git_workdir_returncode in [0,1]
    if git_workdir_returncode == 1:
        print('=> git workdir has changes')
        print('=> please either revert or stage them')
        sys.exit(1)

    pypi_url = 'https://pypi.python.org/pypi/kpa/json'
    latest_version = json.loads(urlopen(pypi_url).read())['info']['version']
    # Note: it takes pypi a minute to update the API, so this can be wrong.
    if latest_version == version:
        new_version_parts = version.split('.')
        new_version_parts[2] = str(1+int(new_version_parts[2]))
        new_version = '.'.join(new_version_parts)
        print('=> autoincrementing version {version} -> {new_version}'.format(version=version, new_version=new_version))
        Path('kpa/version.py').write_text("version = '{new_version}'\n".format(new_version=new_version))
        version = new_version
        subprocess.run(['git','stage','kpa/version.py'])

    git_index_returncode = subprocess.run('git diff-index --quiet --cached HEAD'.split()).returncode
    assert git_index_returncode in [0,1]
    if git_index_returncode == 1:
        print('=> git index has changes')
        subprocess.run(['git','commit','-m',version])

    if not Path('~/.pypirc').expanduser().exists():
        print('=> warning: you need a ~/.pypirc')

    if Path('dist').exists() and list(Path('dist').iterdir()):
        setuppy = Path('dist').absolute().parent / 'setup.py'
        assert setuppy.is_file() and 'kpa' in setuppy.read_text()
        for child in Path('dist').absolute().iterdir():
            assert child.name.startswith('Kpa-'), child
            print('=> unlinking', child)
            child.unlink()

    subprocess.run('python3 setup.py sdist bdist_wheel'.split(), check=True)
    subprocess.run('twine upload dist/*'.split(), check=True)

    if git_index_returncode == 1:
        print('=> Now do `git push`.')
    sys.exit(0)


setup(
    name='Kpa',
    version=version,
    description="<forthcoming>",
    author="Peter VandeHaar",
    author_email="pjvandehaar@gmail.com",
    url="https://github.com/pjvandehaar/kpa",
    classifiers=[
        'Programming Language :: Python :: 3 :: Only',
        'Operating System :: Unix',
    ],

    packages=['kpa'],
    entry_points={'console_scripts': [
        'kpa=kpa.command_line:main',
    ]},
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.4",
    setup_requires=[
        'pytest-runner~=4.2',
    ],
    install_requires=[
        'boltons~=20.2',
    ],
    tests_require=[
        'pytest~=4.0',
    ],
)




import functools
from typing import List, Callable, Any, Iterable, TypeVar, Tuple, Dict

T = TypeVar('T')
T2 = TypeVar('T2')

def assign(func: Callable[..., T]) -> T: return func()
def assign_list(func): return list(func()) # TODO: prefer `@assign @list_from_iter`, which also allows cacheing
def assign_dict(func): return dict(func())

def only_once(func: Callable[..., T]) -> Callable[..., T]:
    already_ran = [False] # TODO: use `nonlocal` or `global` or whatever to avoid `[0]`
    @functools.wraps(func)
    def f(*args, **kwargs):
        if not already_ran[0]:
            already_ran[0] = True
            return func(*args, **kwargs)
    return f

def list_from_iter(func: Callable[..., Iterable[T]]) -> Callable[..., List[T]]:
    @functools.wraps(func)
    def f(*args, **kwargs):
        return list(func(*args, **kwargs))
    return f

def dict_from_iter(func: Callable[..., Iterable[Tuple[T, T2]]]) -> Callable[..., Dict[T, T2]]:
    @functools.wraps(func)
    def f(*args, **kwargs):
        return dict(func(*args, **kwargs))
    return f

# Too complex to typecheck
def apply_to_result(*result_wrapper_funcs):
    def wrap(func):
        @functools.wraps(func)
        def f(*args, **kwargs) -> Any:
            result = func(*args, **kwargs)
            for result_wrapper_func in result_wrapper_funcs:
                result = result_wrapper_func(result)
            return result
        return f
    return wrap
